/* DIMMER2 for 230VAC                             */
/*                                                */
/*                __    _                         */                     
/*     __________/ /_  (_)________   ___  __  __  */ 
/*    / ___/ ___/ __ \/ / ___/ __ \ / _ \/ / / /  */
/*   (__  ) /__/ / / / / /  / /_/ //  __/ /_/ /   */
/*  /____/\___/_/ /_/_/_/  / .___(_)___/\__,_/    */
/*                        /_/                     */
/*                                                */
/* (c) C.Schirp  released for non-commercial use  */
/* IR-routines copyright by Jos van Eijndhoven    */
/*                                                */
/* V1.0  Erstausgabe                              */
/* V1.1  Schnellere Zndentscheidung ohne Jitter  */
/*       Standby-Taste, Optimierungen             */
/* V1.11 Portchange fuer Eagle-PCB                */ 
/*                                                */


#include <d:\projekte\picsel\cc5\16F88.h>
#include <d:\projekte\picsel\cc5\int16CXX.h>
#include <d:\projekte\picsel\cc5\inline.h>

//#include <C:\Eigene_Dateien\pic\cc5x\16F88.h>
//#include <C:\Eigene_Dateien\pic\cc5x\int16CXX.h>
//#include <C:\Eigene_Dateien\pic\cc5x\inline.h>

#pragma config = 0b.0010.1111.0000.0110 

// 13 CP: Flash Program Memory Code Protection bits 1 = Code protection off
// 12 CCPMX: CCP1 Pin Selection bit 1 = CCP1 function on RB0 0 = CCP1 function on RB3
// 11 DEBUG: In-Circuit Debugger Mode bit 1 = In-Circuit Debugger disabled, RB6 and RB7 are GPIO
// 10-9 WRT<1:0>: Flash Program Memory Write Enable bits 11 = Write protection off
//  8 CPD: Data EE Memory Code Protection bit 1 = Code protection off
//  7 LVP: Low-Voltage Programming Enable bit 0 = RB3 is digital I/O, HV on MCLR must be used for programming
//  6 BOREN: Brown-out Reset Enable  1 = BOR enabled 0 = BOR disabled
//  5 MCLRE: RA5/MCLR/VPP Pin Function Select bit 0 = RA5/MCLR/VPP pin function is digital I/O, MCLR internally tied to VDD
//  3 PWRTEN: Power-up Timer Enable bit 1 = PWRT disabled 0 = PWRT enabled
//  2 WDTEN: Watchdog Timer Enable bit 1 = WDT enabled 0 = WDT disabled
//  4, 1-0 FOSC<2:0>: Oscillator Selection bits 
//             100 = INTRC oscillator; port I/O function on both RA6/OSC2/CLKO pin and RA7/OSC1/CLKI pin
//             010 = HS oscillator

#pragma config_reg2 = 0b.1111.1111.1111.1100 
// 13-2 Unimplemented: Read as 1
// 1 IESO: Internal External Switchover bit 1 = Internal External Switchover mode enabled, 0 =  disabled
// 0 FCMEN: Fail-Safe Clock Monitor Enable bit 1 = Fail-Safe Clock Monitor enabled, 0 = disabled



/* Defines */

#define LED_OFF PORTA.1
#define BUZZER PORTA.0
#define ZERO_CROSS PORTB.3
#define GATE PORTB.1

#define MIN_DIM 0	/* minimaler Dimgrad */
#define MAX_DIM 8	/* maximaler Dimgrad */

#define GATE_AKTIV 1
#define GATE_PASSIV 0

#define SAVETIME	625	/* nach SAVETIME mal 8ms wird der dimwert ins eeprom geschrieben */

unsigned char dim_grade;	 //debounced and inverted dimming grade 10...100:
unsigned char dim_1;
unsigned char dim_2;

unsigned char cycles;
unsigned char Puls;		// Statemachine fuer Gatezuendung
unsigned char PulseOn:1;	// Auf eins wenn der nchste Puls gezndet werden soll
unsigned char Stable:1;		// Ist Zerocross-Erkennung stabil
unsigned char actPattern;
unsigned char PatternPos;
unsigned char PatternMask;		// wandernde Bitmaske fr das Firepattern

unsigned char seconds;		// Uhr-Variablen
unsigned char minutes;
unsigned char hours;


uns16  ircode;
uns16  save_counter;
unsigned char irled; // 1 if light is on
unsigned char old_dim_grade;


unsigned char tmr0cpy;



//bit Monitor; // Shadow fuer PB6

bit irledn     @ PORTB.0;

const unsigned char FirePattern[9] = {
0, 
0b1000.0000,
0b1000.1000,
0b1001.0010,
0b1010.1010,
0b1010.1110,
0b1110.1110,
0b1111.1110,
0b1111.1111,
};


/*------------------------------------------------------------------*/
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*------------------------------------------------------------------*/


#pragma origin 4
interrupt int_server( void)
{
    int_save_registers    // W, STATUS (and PCLATH)

    if (INTF)
    {
        /* RB0/INT interrupt */
        irled.0 = !irledn;
        INTF = 0;
        if (irled)     // request interrupt on next edge (up or down)
           INTEDG = 1; // if 1, then interrupt on rising edge
        else
           INTEDG = 0; // is part of OPTION_REG, is 1 on reset

        /* finish and restart TMR0 */
        tmr0cpy = TMR0;
        TMR0 = 0;
        T0IF = 1; // end polling wait-loop, similar to TRM0 wraparound
    }
    
    if (CCP1IF) // capture compare Int -> Halbwellenanfang
    {
 		
	if (PulseOn && Stable) // wenn der Puls aktiv sein soll, Timing starten
 	{
	 	GATE = GATE_AKTIV;
		Puls=1;
		
		TMR2ON=0; // Timer 2 stoppen
	    	TMR2 = 0; // neu aufziehen
		PR2=   23; // ca. 200us
		
		TMR2IE = 1; // Timer 2 Int enablen
		TMR2ON = 1; // Timer 2 starten
	}
	
	if (ZERO_CROSS) // wenn positive Halbwelle trigger auf fallend
	{
		CCP1CON = 0b0000.0100; // Capture every falling edge
		cycles++;
		
		// == Gate-Freigabe berechnen
		PatternPos++;
						
		if (PatternPos > 7)
		{
			PatternPos = 0;
			PatternMask =   1;
		}
		else
		{
			PatternMask<<=1;
		}
		
		if (actPattern & PatternMask)
		{
			PulseOn = 1;
		}
		else
		{
			PulseOn = 0;
		}
			
		
	}
	else
	{
		CCP1CON = 0b0000.0101; // Capture every rising edge
	}
	
	CCP1IF = 0; // CCP-Int-Flag loeschen
	
    }
    

    if (TMR2IF) // Timer 2 Flag, Gatepuls-Erzeugung
    {

    	switch (Puls)
    	{
    		
    		case 1:	GATE = GATE_PASSIV;
    			Puls=2;
    			PR2= 100;//  800us 
    			break; 
    		case 2:	GATE = GATE_AKTIV;
    			Puls=3;
    			PR2= 12;//  100us
    			break; 
    		case 3:	GATE = GATE_PASSIV;
    			Puls=0;
    			TMR2ON = 0; // Timer 2 stoppen
    			TMR2IE = 0; // Timer 2 int disable
    			break; 
    			   			
    		default: GATE = GATE_PASSIV;
    			Puls=0;
    			TMR2ON = 0; // Timer 2 stoppen
    			TMR2IE = 0; // Timer 2 int disable
    			break;
    	}
	TMR2IF =0; // Int-Flag loeschen
    }


   /*
     NOTE: GIE is AUTOMATICALLY cleared on interrupt entry and set
           to 1 on exit (by RETFIE). Setting GIE to 1 inside the
           interrupt service routine will cause nested interrupts
           if an interrupt is pending. Too deep nesting may crash
           the program !
   */

    int_restore_registers // W, STATUS (and PCLATH)
}

/*------------------------------------------------------------------*/
/*   II  N   N  II TTTTT                                            */
/*   II  NN  N  II   T                                              */
/*   II  N N N  II   T                                              */
/*   II  N  NN  II   T                                              */
/*   II  N   N  II   T                                              */
/*------------------------------------------------------------------*/


void init ()
{

	WDTCON = 0b0001.0011;	// Watchdog-Zyklus 1:16394 = 0,5s
	
	PORTB = 0b0000.0000;	// Gate auf Inaktiv
	PORTA = 0b0000.0010;	// LED aus	
	
	ANSEL = 0;	// Alle Analogports auf Digitalbetrieb	
	TRISB = 0b1100.1001; 	// Alles Ausgnge - RB0/Int, ICSPs, ZeroCross Eingang	
	TRISA = 0b1111.0000;	// alles Ausgnge

	/* Timer 0: IR-Empfang */
	/* wird mit OPTION_REG programmiert */

	/* Timer 1: Capture - Messen der Halbwellendauer am Zero-Cross-Eingang RB3/CCP1 */
	/* Aufloesung: 1us */
	
	T1CON = 	0b0001.0000; // Prescaler 1:2, internal Clock, Timer stopped (TMR1ON = 0)
	CCP1CON = 	0b0000.0100; // Capture every falling edge
	
	/* Timer 2: Gatepuls-Timing */
	/* 8us Aufloesung */
	
	T2CON = 	0b0000.0010; // no postscale, timer off, prescale 1:16 


	/*sonstiges*/
	OPTION_REG = 0b.1000.0101; /* no pull-up portB, falling edge int, no WDT prescaler, prescale=64 */


	/* Variablen initialisieren */

	save_counter = 	0; 	// nix schreiben
	irled = 	0;
	seconds = 	0;	// Uhr-Variablen
	minutes =	0;
	hours =		0;
	

	/* Interrupts */
	INTCON = 0; 		// reset and disable all interrupts by default
	PEIE = 1;		// Peripherie-Interrupts freigeben
	INTE = 	1;		// externen Int freigeben (INT0IE)
	
	GIE = 	 1;		// Enable Interrupts
		

}

/*------------------------------------------------------------------*/
/*    EEEE  EEEE  PPPP   RRRR    OOO   M     M                      */
/*    E     E     P   P  R   R  O   O  MM   MM                      */
/*    EEE   EEE   PPPP   RRRR   O   O  M M M M                      */
/*    E     E     P      R R    O   O  M  M  M                      */
/*    EEEE  EEEE  P      R  R    OOO   M     M                      */
/*------------------------------------------------------------------*/



void write_eeprom( unsigned char adr, unsigned char data)
{
	bit prev_gie = GIE;
	GIE = 0; 		// disable interrupts
	EEIF = 0; 		// clear eprom interrupt flag

	while (WR)
		; 			// wait if previous write is still busy

	EEADR = adr; 	// valid range is from 00h to FFh (256 bytes)
	//EEADR &= 0x7f; 	// prevent bugs to overwrite program memory
	EEPGD = 0;	// access data memory
	EEDATA = data;
	WREN = 1;
	EECON2 = 0x55;
	EECON2 = 0xAA;
	WR = 1; 		// start the actual write

	WREN = 0; 		// protect against unintended writes
	GIE = prev_gie; // restore interrupts
}

unsigned char read_eeprom( unsigned char adr)
{
	EEADR = adr; 	// valid range is from 00h to FFh (256 bytes)
	EEPGD = 0;	// access data memory

	//EEADR &= 0x7f; 	// prevent bugs to exceed program memory
	RD = 1;
	return EEDATA;
}

/*------------------------------------------------------------------*/
/*    K  K  EEEE  Y   Y   SSS                                       */
/*    K K   E      Y Y   S                                          */
/*    KK    EE      Y     SSS                                       */
/*    K K   E       Y        S                                      */
/*    K  K  EEEE    Y     SSS                                       */
/*------------------------------------------------------------------*/


void restore_state(void)
{
	unsigned char tester;

	dim_grade = read_eeprom(0);
	dim_1 = read_eeprom(1);
	dim_2 = read_eeprom(2);

	// 2 aus 3
	if (dim_grade==dim_1)
	{
	}
	else if (dim_grade==dim_2)
	{
	}
	else if (dim_1 == dim_2)
	{
		dim_grade = dim_1;
	}

	// Gueltigkeit pruefen
	if (dim_grade < MIN_DIM)
	{
		dim_grade = MIN_DIM;
	}
	else if (dim_grade > MAX_DIM)
	{
		dim_grade = MAX_DIM;
	}
	
	old_dim_grade = dim_grade;	// alten dimgrad setzen, damit er nicht ins EEPROM geschrieben wird
	
	actPattern = FirePattern[dim_grade]; // Aktuelles Gate-Firepattern laden
	
}

/*------------------------------------------------------------------*/
/*    RRR    CCC    55555                                           */
/*    R  R  C   C   5                                               */
/*    RRR   C       5555                                            */
/*    R R   C   C       5                                           */
/*    R  R   CCC    5555                                            */
/*------------------------------------------------------------------*/

// 8MHz crystal, 2MHz cycle time, TMR0 prescale=64: timer units of 32 usec.
// TRM0 expires at 8 msec.

extern uns8 tmr0wait( void)
{
	tmr0cpy = 0xff; // potential longest waiting time, reduced by irled
	while (!T0IF) // wait for TMR0 wrap-around or IRled
		;

	T0IF = 0;
	return tmr0cpy;
}

void ir_bogus( void)
{
	// just eat irled input, as long as we have fast pulses
	uns8 t;

	t = 0;
	while (t < 200)
	   t = tmr0wait();

	ircode = 0xffffL; // bogus code;
}

void ir_philips_rc5(void) // now assuming no 'extended' RC5, so S2==1
{
	// 1 is sequence dark,light:  2x 900us == 2x28
	// 0 is sequence light,dark:  2x 900us == 2x28
	// frame is: one,one,toggle,11xdata
	unsigned char t;
	unsigned char i;

	t = 0;
	i = 0;
	ircode = 0L;
	// starting now in the 2nd (light) half S2
	while (t < 80)
	{
	   // last half of current bit still to be measured, its timer is running
	   ircode <<=1; // insert current value
	   if (irled) // irled input is active low
		   ircode.0 = 1;
	   i++;

	   t = tmr0wait();

	   if (t < 38) // new codebit is same as current, edge at bit boundary
	   {
	   	  t=tmr0wait(); // skip first half of new bit
	   }
	}

	ircode.high8 &= 0x07; // clear start bits S2 and T

	if (i < 13)
	{
		ircode = 0xffffL; // bogus
	}
}


void get_ir( void)
{
	unsigned char ton;
	unsigned char toff;

    ton = tmr0wait(); // measure first light
	toff = tmr0wait(); // following first dark

    if (ton  > 20 && ton  < 36 &&  // nominal 900/32 = 28
        toff > 20 && toff < 36)  // nominal 900/32 = 28
    {
		ir_philips_rc5();
	}
	else
	{
		ir_bogus();
	}
}

void do_remote(void)
{
#pragma updateBank 0

	unsigned char h;

	// systemadresse testen
	h = ircode.high8 & 0b.0000.0111;
	if (h != 4)
	{
		return;
	}

	h = ircode.low8 & 0b.1100.0000;
	if (h != 0)
	{
		return;
	}

	// okay, Kommando gilt uns: Kommando extrahieren
	h= ircode.low8 & 0b.0011.1111;

	// Quellenwahltasten = 1..8
	if (h<=MAX_DIM)
	{
		dim_grade = h;
	}
	

	if (h == 16) // Volume +
	{
		if (dim_grade < MAX_DIM)
		{
			dim_grade++;
		}
	}
	else if (h == 17) // Volume -
	{
		if (dim_grade >= 1)
		{
			dim_grade--;
		}
		
	}
	else if (h == 12) // Standby-> Dimmgrad 0
	{
		dim_grade=0;		
	}

	actPattern = FirePattern[dim_grade]; // Aktuelles Gate-Firepattern laden
#pragma updateBank 1
}

// einen Piepton ausgeben
void beep (void)
{
#pragma updateBank 0

	unsigned char waiter;
	unsigned char counter;	
	
	for (counter = 0; counter < 200; counter++)
	{
		for (waiter=0; waiter < 24 ; waiter++)
		{
			NOP; // 80 = 1.25kHz 24=4kHz
		}
		BUZZER ^= 1;
	}
	
	BUZZER = 0;
#pragma updateBank 1
}

// --- Neu starten der Gate-Ansteuerung nach Reset oder IR-Empfang
void neustart(void)
{
	// Timer 1 auf die nchste fallende Flanke stellen
	CCP1CON = 	0; // Capture stoppen und reset
	TMR1ON = 0; // Timer stoppen
	TMR1H = 0;
	TMR1L = 0;
	CCP1CON = 	0b0000.0100; // Capture every falling edge
	
	// Statemachine auf Anfang
	Puls = 		0;
	
	// Sonstige Variablen
	cycles = 	0;
	PulseOn = 	0;
	Stable =        0;
	PatternPos =	0;
	PatternMask =   1;
	actPattern = FirePattern[dim_grade]; // Aktuelles Gate-Firepattern laden
	
		
	//Timer 1 starten 
	TMR1ON = 1;
	
	// Capture-Interrupt aktivieren
	CCP1IE = 1;	
}

void uhr_betreiben(void)
{
#pragma updateBank 0
	
	if (cycles > 49) // mehr als 50 Cycles-> eine sekunde vergangen
	{
		cycles-=50; // restcycles nicht wegwerfen
		seconds++;
		
		if (seconds > 59) // sechzig sec vergangen -> eine Minute ist um
		{
			seconds = 0;
			minutes++;
			
			if (minutes > 59)
			{
				minutes=0;
				hours++;
			}
		
		}
	}
#pragma updateBank 1	
}



/*------------------------------------------------------------------*/
/*     M     M    A    II  N   N                                    */
/*     MM   MM   A A   II  NN  N                                    */
/*     M M M M  A   A  II  N N N                                    */
/*     M  M  M  AAAAA  II  N  NN                                    */
/*     M     M  A   A  II  N   N                                    */
/*------------------------------------------------------------------*/

void main (void)
{
	unsigned char tmp;

	init();
	
	clrwdt();
	
	beep();

	restore_state();
	
	neustart();

	while (1)
	{
		
		clrwdt();
		
		uhr_betreiben();
		
		// == Betriebsart-LED
		// Dimgrad 0:   alle 10s blink
		// Dimgrad 1-8: sekundentakt-blink
		//
		if (dim_grade)
		{				
			tmp = seconds % 2;
		}
		else
		{
			tmp = seconds % 10;
		}
						
		if (tmp)
		{
			LED_OFF = 1;
		}
		else
		{
			LED_OFF = 0;
		}
				
		// == Timer-Jobs behandeln
		
		if (seconds > 3) // Starttimer zur Stabilisierung
		{
			Stable = 1;
		}
		
		if (hours > 7) // Notbremse ziehen: nach 8 Stunden Dimmgrad auf 0
		{
			dim_grade = 0;
			seconds = 0; // Zeitrechnung neu starten
			minutes = 0;
			hours = 0;	
		}
		
		
		tmp = tmr0wait(); // wrap-around or interrupt, wrap-around takes 8ms
	
	    	if (irled) //came out of wait by interrupt: light on, start of new code
		{
			get_ir();
			do_remote();
		}
		
		// wenn sich der dimmgrad gendert hat, ins EEPROM schreiben
		if (old_dim_grade != dim_grade)
		{
			beep();
			
			save_counter = SAVETIME;
			
			old_dim_grade = dim_grade;			
		}
		
		// wartezeit vor dem speichern abgelaufen?
		if (save_counter == 1)
		{	
			write_eeprom(0,dim_grade);
			write_eeprom(1,dim_grade);
			write_eeprom(2,dim_grade);
			
			save_counter = 0; // speichern erledigt
		}
		else if (save_counter > 1)
		{
			save_counter--;
		}
		

	} /* while (1) */
}